"""Input-output for PyConTurb files
"""
import os
from struct import pack, unpack
import time

import numpy as np
import pandas as pd

from pyconturb._version import __version__


_HAWC2_BIN_FMT = '<f'  # HAWC2 binary turbulence datatype
_TS_BIN_FMT = '<h4l12fl'  # TurbSim binary format


def bts_to_df(path):
    """Load TurbSim-style .bts file to pyconturb dataframe"""
    if path.endswith('.bts'):  # remove file extension, will be added back later
        path = path[:-4]
    u_scl = np.zeros(3, np.float32)
    u_off = np.zeros(3, np.float32)
    with open(path + '.bts', 'rb') as fl:
        (_,  # periodic (7) or non periodic (8), unused
         nz,
         ny,
         _,  # no. tower ptsbelow box, unused
         nt,
         dz,
         dy,
         dt,
         uhub,
         zhub,
         z0,
         u_scl[0],
         u_off[0],
         u_scl[1],
         u_off[1],
         u_scl[2],
         u_off[2],
         strlen) = unpack(_TS_BIN_FMT, fl.read(70))
        desc_str = fl.read(strlen)  # skip these bytes
        nbt = 3 * ny * nz * nt
        dat = np.rollaxis(np.frombuffer(fl.read(2 * nbt), dtype=np.int16).astype(
            np.float32).reshape([3, ny, nz, nt], order='F'), 2, 1)  # 3 x ny x nz x nt
    dat -= u_off[:, None, None, None]
    dat /= u_scl[:, None, None, None]
    # create pyconturb-flavor dataframe
    t = np.arange(nt) * dt
    turb_df = pd.DataFrame(index=t)
    for i, c in enumerate('uvw'):
        arr = np.transpose(dat[i], (2, 0, 1))  # nt x ny x nz
        comp_df = pd.DataFrame(arr.reshape(nt, ny*nz),
                               index=t).add_prefix(f'{c}_p')
        turb_df = turb_df.join(comp_df, how='outer')
    return turb_df


def df_to_bts(turb_df, spat_df, path, uzhub=None, periodic=True):
    """pyconturb-style turbulence dataframe to TurbSim-style binary file
    Code modified based on `turbsim` in PyTurbSim:
        https://github.com/lkilcher/pyTurbSim/blob/master/pyts/io/write.py

    Notes
    -----
    * The turbulence must have been generated on a y-z grid.
    * The naming convention must be 'u_p0', 'v_p0, 'w_p0', 'u_p1', etc.,
        where the point indices proceed vertically along z before horizontally
        along y.
    """
    if path.endswith('.bts'):  # remove file extension, will be added back later
        path = path[:-4]
    # format-specific constants
    intmin = -32768  # minimum integer
    intrng = 65535  # range of integers
    # calculate intermediate parameters
    y = np.sort(np.unique(spat_df.loc['y'].values))
    z = np.sort(np.unique(spat_df.loc['z'].values))
    ny = y.size  # no. of y points in grid
    nz = z.size  # no. of z points in grif
    nt = turb_df.shape[0]  # no. of time steps
    if y.size == 1:
        dy = 0
    else:
        dy = np.mean(y[1:] - y[:-1])  # hopefully will reduce possible errors
    if z.size == 1:
        dz = 0
    else:
        dz = np.mean(z[1:] - z[:-1])  # hopefully will reduce possible errors
    dt = turb_df.index[-1] / (turb_df.shape[0] - 1)  # time step
    if uzhub is None:  # default is center of grid
        zhub = z[z.size // 2]  # halfway up
        u_df = turb_df.filter(regex=f'u_', axis=1).values
        uhub = u_df[:, u_df.shape[1] // 2].mean()  # mean of center of grid
    else:
        uhub, zhub = uzhub
    # convert pyconturb dataframe to pyturbsim format (3 x ny x nz x nt)
    ts = np.empty((3, ny, nz, nt))
    for i, c in enumerate('uvw'):
        arr = turb_df.filter(regex=f'{c}_', axis=1).values.reshape((nt, ny, nz))
        ts[i] = np.transpose(arr, (1, 2, 0))  # reorder to ny, nz, nt
    # initialize output arrays
    u_minmax = np.empty((3, 2), dtype=np.float32)  # mins and maxs of time series
    u_off = np.empty((3), dtype=np.float32)  # offsets of each time series
    u_scl = np.empty((3), dtype=np.float32)  # scales of each time series
    desc_str = 'generated by PyConTurb v%s, %s.' % (
        __version__,
        time.strftime('%b %d, %Y, %H:%M (%Z)', time.localtime()))  # description
    # calculate the scales and offsets of each time series
    out = np.empty((3, ny, nz, nt), dtype=np.int16)
    for ind in range(3):
        u_minmax[ind] = ts[ind].min(), ts[ind].max()
        if np.isclose(u_minmax[ind][0], u_minmax[ind][1]):
            u_scl[ind] = 1
        else:
            u_scl[ind] = intrng / np.diff(u_minmax[ind])
        u_off[ind] = intmin - u_scl[ind] * u_minmax[ind, 0]
        out[ind] = (ts[ind] * u_scl[ind] + u_off[ind]).astype(np.int16)
    with open(path + '.bts', 'wb') as fl:
        # write the header
        fl.write(pack(_TS_BIN_FMT,
                      [7, 8][periodic],  # 7 is not periodic, 8 is periodic
                      nz,
                      ny,
                      0,  # assuming 0 tower points below grid
                      nt,
                      dz,
                      dy,
                      dt,
                      uhub,
                      zhub,
                      z[0],
                      u_scl[0],
                      u_off[0],
                      u_scl[1],
                      u_off[1],
                      u_scl[2],
                      u_off[2],
                      len(desc_str)))
        fl.write(desc_str.encode(encoding='UTF-8'))
        # Swap the y and z indices so that fortran-order writing agrees with the file format.
        # Also, we swap the order of z-axis to agree with the file format.
        # Write the data so that the first index varies fastest (F order).
        # The indexes vary in the following order:
        # component (fastest), y-index, z-index, time (slowest).
        fl.write(np.rollaxis(out, 2, 1).tobytes(order='F'))


def df_to_h2turb(turb_df, spat_df, path, prefix=''):
    """pyconturb-style turbulence dataframe to binary files for hawc2

    Notes
    -----
    * The turbulence must have been generated on a y-z grid.
    * The naming convention must be 'u_p0', 'v_p0, 'w_p0', 'u_p1', etc.,
       where the point indices proceed vertically along z before horizontally
       along y.
    """
    nx = turb_df.shape[0]  # turbulence dimensions for reshaping
    ny = len(set(spat_df.loc['y'].values))
    nz = len(set(spat_df.loc['z'].values))
    # make and save binary files for all three components
    for c in 'uvw':
        arr = turb_df.filter(regex=f'{c}_', axis=1).values.reshape((nx, ny, nz))
        bin_path = os.path.join(path, f'{prefix}{c}.bin')
        with open(bin_path, 'wb') as bin_fid:
            arr.astype(np.dtype(_HAWC2_BIN_FMT)).tofile(bin_fid)
    return


def h2turb_to_arr(spat_df, path):
    """Raw-load a hawc2 turbulent binary file to float-64 numeric array"""
    ny, nz = pd.unique(spat_df.loc['y']).size, pd.unique(spat_df.loc['z']).size
    bin_arr = np.fromfile(path, dtype=np.dtype(_HAWC2_BIN_FMT))
    nx = bin_arr.size // (ny * nz)
    if (nx * ny * nz) != bin_arr.size:
        raise ValueError('Binary file size does not match spat_df!')
    bin_arr.shape = (nx, ny, nz)
    return bin_arr


def h2turb_to_df(spat_df, path, nt=600, dt=1, prefix=''):
    """load a hawc2 binary file into a float-64 pandas datafram with transform to uvw"""
    t = np.arange(nt) * dt
    turb_df = pd.DataFrame(index=t)
    for c in 'uvw':
        comp_path = os.path.join(path, f'{prefix}{c}.bin')
        arr = h2turb_to_arr(spat_df, comp_path)
        nx, ny, nz = arr.shape
        comp_df = pd.DataFrame(arr.reshape(nx, ny*nz),
                               index=t).add_prefix(f'{c}_p')
        turb_df = turb_df.join(comp_df, how='outer')
    turb_df = turb_df[[f'{c}_p{i}' for i in range(ny*nz) for c in 'uvw']]
    return turb_df



# -*- coding: utf-8 -*-
"""Defining, calculating, and saving/loading spatial coherence values.
"""
import itertools
import os

import h5py
import numpy as np
from scipy.linalg import cholesky

from pyconturb._utils import (check_chunk_idcs, combine_spat_con, get_chunk_idcs,
                              get_freq, message)


_HDF5_DSNAME = 'coherence'  # dataset name for HDF5 coherence files


def calculate_coh_mat(freq, spat_df, con_tc=None, coh_model='iec', dtype=np.float64,
                      chunk_idcs=None, **kwargs):
    """Calculate coherence matrix using Cholesky decomposition.
    
    If constraint is passed in, the first ``n_c`` rows/columns of the returned array
    correlate the constraints and the subsequent ``n_sp`` rows/columns correspond to the 
    requested simulation points.

    Parameters
    ----------
    freq : array-like
        [Hz] Full frequency vector for coherence calculations. Option to calculate 
        coherence for a subset using ``chunk_idcs`` keyword argument. Dimension is
        ``(n_f,)``.
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have rows ``[k, x, y, z]``,
        and each of the ``n_sp`` columns corresponds to a different spatial location and
        turbine component (u, v or w).
    con_tc : pyconturb TimeConstraint, optional
        Optional constraining data for the simulation. The TimeConstraint object is built
        into PyConTurb; see documentation for more details. Default is none (no 
        constaint). Containts ``n_c`` columns, one for each constraint location.
    coh_model : str, optional
        Spatial coherence model specifier. Default is "iec" (IEC 61400-1, Ed. 4).
    dtype : data type, optional
        Change precision of calculation (np.float32 or np.float64). Will reduce the 
        storage, and might slightly reduce the computational time. Default is np.float64.
    chunk_idcs : int or numpy.array, optional
        Indices of ``freq`` for which coherence should be calculated or loaded. Dimension
        is ``(n_fchunk,)`` if given. Default is None (get all frequencies in ``freq``).
    **kwargs
        Keyword arguments to pass into ``get_iec[3d]_cor_mat``.


    Returns
    -------
    coh_mat : numpy.ndarray
        Generated coherence matrix. Dimension is ``(n_fchunk, n_c+n_sp, n+n_sp)``.
    """
    
    # preppring
    freq = np.atleast_1d(freq)  # make freqs array if not
    all_spat_df = combine_spat_con(spat_df, con_tc)  # combine cons and sim
    chunk_idcs = check_chunk_idcs(freq, chunk_idcs)  # update chunk_idcs if not given
    
    # throw error if chunk indices not a numpy array
    if type(chunk_idcs) not in [int, np.ndarray]:
        raise ValueError('chunk_idcs must be integer or numpy array!')
    
    # get chunk frequency array
    freq_chunk = freq[chunk_idcs]

    # get CORRELATION matrix based on coherence model
    if coh_model == 'iec':  # IEC coherence model
        coh_mat = get_iec_cor_mat(freq_chunk, all_spat_df, dtype=dtype,
                                  coh_model=coh_model, **kwargs)
    elif coh_model == 'iec3d':  # IEC coherence model, but in u, v, and w
        coh_mat = get_iec3d_cor_mat(freq_chunk, all_spat_df, dtype=dtype,
                                    coh_model=coh_model, **kwargs)
    else:  # unknown coherence model
        raise ValueError(f'Coherence model "{coh_model}" not recognized!')

    # get COHERENCE matrix (in-place) via cholesky decomp. In-place saves memory.
    # TODO! EASY parallelization here
    for i in range(coh_mat.shape[0]):
        coh_mat[i] = cholesky(coh_mat[i], check_finite=False, lower=True)

    return coh_mat



def chunker(iterable, nPerChunks):
    """Return list of nPerChunks elements of an iterable"""
    it = iter(iterable)
    while True:
       chunk = list(itertools.islice(it, nPerChunks))
       if not chunk:
           return
       yield chunk


def generate_coherence_file(spat_df, coh_file, freq=None, con_tc=None, coh_model='iec',
                            nf_chunk=1, dtype=np.float64, verbose=False, **kwargs):
    """Calculate a coherence matrix and save it to an HDF5 file for later reuse.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have rows ``[k, x, y, z]``,
        and each of the ``n_sp`` columns corresponds to a different spatial location and
        turbine component (u, v or w).
    coh_file : str or pathlib.Path
        Path to file from which to load coherence. Assumed to be an HDF5 file with
        dataset "coherence" containing a 2D coherence array with dimensions 
        ``(n_f, (n_c+n_sp)^2)``.
    con_tc : pyconturb TimeConstraint, optional
        Optional constraining data for the simulation. The TimeConstraint object is built
        into PyConTurb; see documentation for more details. Default is none (no 
        constaint). Containts ``n_c`` columns, one for each constraint location.
    freq : array-like, optional
        [Hz] Full frequency vector for coherence calculations if con_tc not given. 
        Option to only calculate coherence for a subset of ``freq`` using ``chunk_idcs``
        keyword argument. Dimension of ``freq`` is is ``(n_f,)``.
    coh_model : str, optional
        Spatial coherence model specifier. Default is "iec" (IEC 61400-1, Ed. 4).
    dtype : data type, optional
        Change precision of calculation (np.float32 or np.float64). Will reduce the 
        storage, and might slightly reduce the computational time. Default is np.float64.
    nf_chunk : int, optional
        Number of frequencies in a chunk of analysis. Increasing this number may speed
        up computation but may result in more (or too much) memory used. Smaller grids
        may benefit from larger values for ``nf_chunk``. Default is 1.
    verbose : boolean, optional
        Whether to print updates to screen during calculation. Default is False.
    **kwargs
        Keyword arguments to pass into ``calculate_coh_mat``.


    Returns
    -------
    coh_mat : numpy.ndarray
        Generated coherence matrix. Dimension is ``(n_fchunk, n_sp, n_sp)``.
    """
    
    # get freq from con_tc if not given
    if freq is None:
        if con_tc is None:
            raise ValueError('Must given either "freq" or "con_tc" keyword argument!')
        freq = get_freq(T=con_tc.get_T(), nt=con_tc.get_nt())[1]
        

    # sizes of arrays
    n_sp = spat_df.shape[1]  # points to simulate
    n_c = 0 if con_tc is None else con_tc.shape[1]  # constraints
    n_f = np.size(freq)
    n_chunks = np.ceil(n_f / nf_chunk).astype(int)
    
    # open the file
    with h5py.File(coh_file, 'w') as hf:
        
        # initialize dataset
        dset = hf.create_dataset(_HDF5_DSNAME, (n_f, (n_sp + n_c)**2), dtype=dtype)
        
        # for each chunk
        for i_chunk in range(n_chunks):
            message(f'Processing chunk {i_chunk+1}/{n_chunks}...', verbose)
            
            # get the indices of frequencies for this chunk
            chunk_idcs = get_chunk_idcs(freq, i_chunk, nf_chunk)
            
            # calculate the coherence matrix
            chunk_coh_mat = calculate_coh_mat(freq, spat_df, coh_model=coh_model,
                                              dtype=dtype, chunk_idcs=chunk_idcs,
                                              con_tc=con_tc, **kwargs)

            # add coherence to the dataset
            dset[chunk_idcs, :] = chunk_coh_mat.reshape(len(chunk_idcs), (n_sp + n_c)**2)
            
    message(f'DONE. Coherence file saved to {coh_file}.', verbose)
            
    return


def get_coh_mat(freq, spat_df, con_tc=None, coh_model='iec', dtype=np.float64,
                coh_file=None, chunk_idcs=None, **kwargs):
    """Get coherence matrix (either calculate or load) for set of frequencies.
    
    If the ``coh_file`` option is given, this function calls ``load_coh_mat`` to load the
    requested matrix from file. If no file name is given, this function instead calls
    ``calculate_coh_mat`` to calculate the coherence matrix.
    
    Parameters
    ----------
    freq : array-like
        [Hz] Full frequency vector for coherence calculations. Option to calculate 
        coherence for a subset using `chunk_idcs` keyword argument. Dimension is
        ``(n_f,)``.
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have rows ``[k, x, y, z]``,
        and each of the ``n_sp`` columns corresponds to a different spatial location and
        turbine component (u, v or w).
    con_tc : pyconturb TimeConstraint, optional
        Optional constraining data for the simulation. The TimeConstraint object is built
        into PyConTurb; see documentation for more details. Default is none (no 
        constaint). Containts ``n_c`` columns, one for each constraint location.
    coh_model : str, optional
        Spatial coherence model specifier. Default is "iec" (IEC 61400-1, Ed. 4).
    dtype : data type, optional
        Change precision of calculation (np.float32 or np.float64). Will reduce the 
        storage, and might slightly reduce the computational time. Default is np.float64.
    coh_file : str or pathlib.Path, optional
        Path to file from which to load coherence. Assumed to be an HDF5 file with
        dataset "coherence" containing a 2D coherence array with dimensions 
        ``(n_f, n_sp^2)``. Default is None (calculate, don't load from file).
    chunk_idcs : int or numpy.array, optional
        Indices of `freq` for which coherence should be calculated or loaded. Dimension
        is ``(n_fchunk,)`` if given. Default is None (get all frequencies in ``freq``).
    **kwargs
        Keyword arguments to pass into ``calculate_coh_mat``.


    Returns
    -------
    coh_mat : numpy.ndarray
        Generated or loaded coherence matrix. Dimension is
        ``(n_fchunk, n_c+n_sp, n+n_sp)``.
    """
    
    # update chunk_idcs if not given
    chunk_idcs = check_chunk_idcs(freq, chunk_idcs)

    # if filename not given, calculate the coherence matrix
    if coh_file is None:
        coh_mat = calculate_coh_mat(freq, spat_df, coh_model=coh_model, dtype=dtype,
                                    con_tc=con_tc, chunk_idcs=chunk_idcs, **kwargs)
    
    # if a filename IS given
    else:
        
        # raise error if file not existing
        if not os.path.isfile(coh_file):
            raise ValueError(f'File "{coh_file}" not found! Generate it using `generate_coherence_file`.')
        
        # otherwise, load it from the file
        coh_mat = load_coh_mat(coh_file, freq, chunk_idcs=chunk_idcs)

    return coh_mat



def get_iec_cor_mat(freq, spat_df, dtype=np.float64, ed=4, **kwargs):
    """Create IEC 61400-1 Ed. 3/4 correlation matrix for given frequencies. [nf x ns x ns]
    """
    # preliminaries
    if ed not in [3, 4]:  # only allow edition 3
        raise ValueError('Only editions 3 or 4 are permitted.')
    if any([k not in kwargs for k in ['u_ref', 'l_c']]):  # check kwargs
        raise ValueError('Missing keyword arguments for IEC coherence model')
    freq = np.array(freq).reshape(-1, 1)  # nf x 1
    nf, ns = freq.size, spat_df.shape[1]
    # intermediate variables
    yz = spat_df.loc[['y', 'z']].values.astype(float)
    cor_mat = np.repeat(np.eye((ns), dtype=dtype)[None, :, :], nf, axis=0) # TODO use sparse matrix 
    exp_constant = np.sqrt( (1/ kwargs['u_ref'] * freq)**2 + (0.12 / kwargs['l_c'])**2).astype(dtype)  # nf x 1
    i_comp = np.arange(ns)[spat_df.iloc[0, :].values == 0]  # Selecting only u-components
    # loop through number of combinations, nPerChunks at a time to reduce memory impact
    for ii_jj in chunker(itertools.combinations(i_comp, 2), nPerChunks=10000):
        # get indices of point-pairs
        ii = np.array([tup[0] for tup in ii_jj])
        jj = np.array([tup[1] for tup in ii_jj])
        # calculate distances and coherences
        r = np.sqrt((yz[0, ii] - yz[0, jj])**2 + (yz[1, ii] - yz[1, jj])**2)  # nchunk
        coh_values = np.exp(-12 * np.abs(r) * exp_constant)  # nf x nchunk
        # Previous method (same math, different numerics)
        cor_mat[:, ii, jj] = coh_values  # nf x nchunk
        cor_mat[:, jj, ii] = np.conj(coh_values)
    return cor_mat


def get_iec3d_cor_mat(freq, spat_df, dtype=np.float64, **kwargs):
    """Create IEC-flavor correlation but for all 3 components. Returns nf x ns x ns.
    """
    # preliminaries
    if any([k not in kwargs for k in ['u_ref', 'l_c']]):  # check kwargs
        raise ValueError('Missing keyword arguments for IEC coherence model')
    freq = np.array(freq).reshape(-1, 1)  # nf x 1
    nf, ns = freq.size, spat_df.shape[1]
    # intermediate variables
    yz = spat_df.loc[['y', 'z']].values.astype(float)
    cor_mat = np.repeat(np.eye((ns), dtype=dtype)[None, :, :], nf, axis=0) # TODO use sparse matrix 
    for ic in range(3):
        Lc = kwargs['l_c'] * [1, 2.7/8.1 , 0.66/8.1 ][ic]
        exp_constant = np.sqrt( (1/ kwargs['u_ref'] * freq)**2 + (0.12 / Lc)**2).astype(dtype)
        i_comp = np.arange(ns)[spat_df.iloc[0, :].values == ic]  # Selecting only u-components
        # loop through number of combinations, nPerChunks at a time to reduce memory impact
        for ii_jj in chunker(itertools.combinations(i_comp, 2), nPerChunks=10000):
            # get indices of point-pairs
            ii = np.array([tup[0] for tup in ii_jj])
            jj = np.array([tup[1] for tup in ii_jj])
            # calculate distances and coherences
            r = np.sqrt((yz[0, ii] - yz[0, jj])**2 + (yz[1, ii] - yz[1, jj])**2)
            coh_values = np.exp(-12 * np.abs(r) * exp_constant)
            # Previous method (same math, different numerics)
            cor_mat[:, ii, jj] = coh_values  # nf x nchunk
            cor_mat[:, jj, ii] = np.conj(coh_values)
    return cor_mat


def load_coh_mat(coh_file, freq=None, chunk_idcs=None):
    """Load all or part of a coherence matrix from an HDF5 file.
    
    Parameters
    ----------
    coh_file : str or pathlib.Path
        Path to file from which to load coherence. Assumed to be an HDF5 file with
        dataset "coherence" containing a 2D coherence array with dimensions 
        ``(n_f, n_sp^2)``.
    freq : array-like, optional
        [Hz] Full frequency vector that was used in coherence calculations. Omitting
        ``freq`` means load the entire dataset. Option to load a subset of the coherence
        matrix using `chunk_idcs` keyword argument (indexes along frequency). 
        Dimension is ``(n_f,)``.
    chunk_idcs : int or numpy.array
        Indices of `freq` for which coherence should be calculated or loaded. Dimension
        is ``(n_fchunk,)`` if given. Default is None (get all frequencies in `freq`).
    
    
    Returns
    -------
    coh_mat : numpy.ndarray
        Generated coherence matrix. Dimension is ``(n_fchunk, n_sp, n_sp)``.
    """
    
    # load everything if freq not given
    if freq is None:
        with h5py.File(coh_file, 'r') as hf:
            coh_mat_2d = hf[_HDF5_DSNAME][:, :]
        
    else:
        # update chunk_idcs if not given
        chunk_idcs = check_chunk_idcs(freq, chunk_idcs)
        
        # load the slice from memory
        with h5py.File(coh_file, 'r') as hf:
            coh_mat_2d = hf[_HDF5_DSNAME][chunk_idcs, :]
    
    ns = np.sqrt(coh_mat_2d.shape[1]).astype(int)
    
    return coh_mat_2d.reshape((-1, ns, ns))
    



#   version:    1.0     First pass at PyConTurb
#               2.0     Switch to uvw, other changes
#                2.5    Fixed bug in custom profiles
#               2.6     Condata is deprecated.
#               2.7     Write coherence to file
__version__ = '2.7.1'


# -*- coding: utf-8 -*-
"""Create an unconstrained or constrained turbulence box.

The main function function, ``gen_turb`` can be used both with and without
constraining data. Please see the Examples section in the documentation
to see how to use it.
"""
import warnings

import numpy as np
import pandas as pd

from pyconturb.coherence import get_coh_mat
from pyconturb.core import TimeConstraint
from pyconturb.magnitudes import get_magnitudes
from pyconturb.sig_models import iec_sig, data_sig
from pyconturb.spectral_models import kaimal_spectrum, data_spectrum
from pyconturb.wind_profiles import get_wsp_values, power_profile, data_profile
from pyconturb._utils import (combine_spat_con, _spat_rownames, _DEF_KWARGS,
                              clean_turb, check_sims_collocated, get_freq, get_chunk_idcs,
                              message)


def gen_turb(spat_df, T=600, nt=600, con_tc=None, coh_model='iec', coh_file=None,
             wsp_func=None, sig_func=None, spec_func=None,
             interp_data='none', seed=None, nf_chunk=1, dtype=np.float64, verbose=False,
             **kwargs):
    """Generate a turbulence box (constrained or unconstrained).

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have rows `[k, x, y, z]`,
        and each of the `n_sp` columns corresponds to a different spatial location and
        turbine component (u, v or w).
    T : float, optional
        Total length of time to simulate in seconds. Default is 600.
    nt : int, optional
        Number of time steps for generated turbulence. Default is 600.
    con_tc : pyconturb TimeConstraint, optional
        Optional constraining data for the simulation. The TimeConstraint object is built
        into PyConTurb; see documentation for more details.
    coh_model : str, optional
        Spatial coherence model specifier. Default is IEC 61400-1, Ed. 4.
    coh_file : str or pathlib.Path
        Path to file from which to load coherence. Assumed to be an HDF5 file with
        dataset "coherence" containing a 2D coherence array with dimensions 
        ``(n_f, (n_sp + n_c)^2)``. See documentation for more details.
    wsp_func : function, optional
        Function to specify spatial variation of mean wind speed. See details
        in `Mean wind speed profiles` section.
    sig_func : function, optional
        Function to specify spatial variation of turbulence standard deviation.
        See details in `Turbulence standard deviation` section.
    spec_func : function, optional
        Function to specify spatial variation of turbulence power spectrum. See
        details in `Turbulence spectra` section.
    interp_data : str, optional
        Interpolate mean wind speed, standard deviation, and/or power spectra profile
        functions from provided constraint data. Possible options are ``'none'`` (use
        provided/default profile functions), ``'all'`` (interpolate all profile,
        functions), or a list containing and combination of ``'wsp'``, ``'sig'`` and
        ``'spec'`` (interpolate the wind speed, standard deviation and/or power spectra).
        Default is IEC 61400-1 profiles (i.e., no interpolation).
    seed : int, optional
        Optional random seed for turbulence generation. Use the same seed and
        settings to regenerate the same turbulence box.
    nf_chunk : int, optional
        Number of frequencies in a chunk of analysis. Increasing this number may speed
        up computation but may result in more (or too much) memory used. Smaller grids
        may benefit from larger values for ``nf_chunk``. Default is 1.
    verbose : bool, optional
        Print extra information during turbulence generation. Default is False.
    dtype : data type, optional
        Change precision of calculation (np.float32 or np.float64). Will reduce the 
        storage, and might slightly reduce the computational time. Default is np.float64.
    **kwargs
        Optional keyword arguments to be fed into the
        spectral/turbulence/profile/etc. models.

    Returns
    -------
    turb_df : pandas.DataFrame
        Generated turbulence box. Each row corresponds to a time step and each
        column corresponds to a point/component in ``spat_df``.
    """
    message('Beginning turbulence simulation...', verbose)

    # if con_data passed in, throw error
    if 'con_data' in kwargs:
        raise ValueError('The "con_data" option is deprecated and can no longer be used.'
                         ' Please see documentation for updated usage.')
    # if dt passed in, throw deprecation warning
    if 'dt' in kwargs:
        warnings.warn('The "dt" keyword argument is deprecated and will be removed in '
                      + 'future releases. Please use the "nt" argument instead.',
                      DeprecationWarning)
        nt = int(np.ceil(T / kwargs['dt']))  # no. time steps
    # if asked to interpret but no data, throw warning
    if (((interp_data == 'all') or isinstance(interp_data, list)) and (con_tc is None)):
        raise ValueError('If "interp_data" is not "none", constraints must be given!')
    # return None if all simulation points collocated with constraints
    if check_sims_collocated(spat_df, con_tc):
        print('All simulation points collocated with constraints! '
              + 'Nothing to simulate.')
        return None

    dtype_complex = np.complex64 if dtype==np.float32 else np.complex128  # complex dtype

    # add T, nt, con_tc to kwargs
    kwargs = {**_DEF_KWARGS, **kwargs, 'T': T, 'nt': nt, 'con_tc': con_tc}
    wsp_func, sig_func, spec_func = assign_profile_functions(wsp_func, sig_func,
                                                             spec_func, interp_data)

    # assign/create constraining data
    t, freq = get_freq(**kwargs)  # time array
    if con_tc is None:  # create empty constraint data if not passed in
        constrained, n_d = False, 0  # no. of constraints
        con_tc = TimeConstraint(index=_spat_rownames)
    else:
        constrained = True
        n_d = con_tc.shape[1]  # no. of constraints
        if not np.allclose(con_tc.get_time().index.values.astype(float), t):
            raise ValueError('TimeConstraint time does not match requested T, dt!')

    # combine data and sim spat_dfs
    all_spat_df = combine_spat_con(spat_df, con_tc)  # all sim points
    n_s = all_spat_df.shape[1]  # no. of total points to simulate
    n_f = freq.size # no. freqs

    one_point = True if n_s == 1 else False  # only one point, skip coherence

    # get magnitudes of points to simulate. (nf, nsim). con_tc in kwargs.
    # TODO: Change this to similar form as get_coh_mat
    sim_mags = get_magnitudes(all_spat_df.iloc[:, n_d:], spec_func, sig_func, **kwargs)

    if constrained:
        conturb_fft = np.fft.rfft(con_tc.get_time().values, axis=0) / nt  # constr fft
        con_mags = np.abs(conturb_fft)  # mags of constraints
        all_mags = np.concatenate((con_mags, sim_mags), axis=1)  # con and sim
    else:
        all_mags = sim_mags  # just sim
    all_mags = all_mags.astype(dtype, copy=False)

    # get uncorrelated phasors for simulation
    np.random.seed(seed=seed)  # initialize random number generator
    sim_unc_pha = np.exp(1j * 2*np.pi * np.random.rand(n_f, n_s - n_d))
    if not (nt % 2):  # if even time steps, last phase must be 0 or pi for real sig
        sim_unc_pha[-1, :] = np.exp(1j * np.round(np.real(sim_unc_pha[-1, :])) * np.pi)

    # no coherence if one point
    if one_point:
        turb_fft = all_mags * sim_unc_pha

    # if more than one point, correlate everything
    else:
        turb_fft = np.zeros((n_f, n_s), dtype=dtype_complex)
        n_chunks = int(np.ceil(freq.size / nf_chunk))

        # loop through frequencies
        for i_f in range(1, freq.size):
            i_chunk = i_f // nf_chunk  # calculate chunk number
            if (i_f - 1) % nf_chunk == 0:  # genr cohrnc chunk when needed
                message(f'  Processing chunk {i_chunk + 1} / {n_chunks}', verbose)
                
                chunk_idcs = get_chunk_idcs(freq, i_chunk, nf_chunk)
                chunk_coh_mat = get_coh_mat(freq, spat_df, coh_model=coh_model,
                                            chunk_idcs=chunk_idcs, dtype=dtype,
                                            coh_file=coh_file, **kwargs)

            # coherence array for this frequency
            coh_mat = chunk_coh_mat[i_f % nf_chunk]  # ns x ns

            # get cholesky decomposition of coherence matrix, make coherence
            L_mat = coh_mat * all_mags[i_f]

            # if constraints, assign data unc_pha
            if constrained:
                dat_unc_pha = np.linalg.solve(L_mat[:n_d, :n_d], conturb_fft[i_f, :])
            else:
                dat_unc_pha = []
            unc_pha = np.concatenate((dat_unc_pha, sim_unc_pha[i_f, :]))
            cor_pha = L_mat @ unc_pha

            # calculate and save correlated Fourier components
            turb_fft[i_f, :] = cor_pha

        del chunk_coh_mat  # free up memory

    # convert to time domain and pandas dataframe
    turb_arr = np.fft.irfft(turb_fft, axis=0, n=nt) * nt
    turb_arr = turb_arr.astype(dtype, copy=False)
    turb_df = pd.DataFrame(turb_arr, columns=all_spat_df.columns, index=t)

    # return just the desired simulation points
    turb_df = clean_turb(spat_df, all_spat_df, turb_df)

    # add in mean wind speed according to specified profile
    wsp_profile = get_wsp_values(spat_df, wsp_func, **kwargs)
    turb_df[:] += wsp_profile

    message('Turbulence generation complete.', verbose)

    return turb_df


def assign_profile_functions(wsp_func, sig_func, spec_func, interp_data):
    """Assign profile functions based on user inputs"""
    # assign iec profile functions as default if nothing was passed in
    prof_funcs = [wsp_func, sig_func, spec_func]  # given inputs
    iec_funcs = [power_profile, iec_sig,  kaimal_spectrum]  # iec default functions
    prof_funcs = [tup[tup[0] is None] for tup in zip(prof_funcs, iec_funcs)]  # overwrite
    # change interp_data to list format if 'all' or 'none' passed in
    if interp_data == 'none': interp_data = []  # no profs interpolated
    elif interp_data == 'all': interp_data = ['wsp', 'sig', 'spec']  # all profs interp'd
    elif not isinstance(interp_data, list):  # bad input
        raise ValueError('"interp_data" must be either "all", "none", or a list!')
    # assign the interp_data profiles IF no custom function was passed in
    for prof in interp_data:
        if (prof == 'wsp') and (wsp_func is None): prof_funcs[0] = data_profile
        elif (prof == 'sig') and (sig_func is None): prof_funcs[1] = data_sig
        elif (prof == 'spec') and (spec_func is None): prof_funcs[2] = data_spectrum
        else: raise ValueError(f'Unknown profile type "{prof}"!')  # bad input
    return prof_funcs



# -*- coding: utf-8 -*-
"""utility functions
"""
import os

import numpy as np
import pandas as pd
import scipy.interpolate as sciint

from pyconturb.core import TimeConstraint


_spat_rownames = ['k', 'x', 'y', 'z']  # row names of spatial df
_DEF_KWARGS = {'u_ref': 0, 'z_ref': 90, 'alpha': 0.2, 'turb_class': 'A',
               'l_c': 340.2}  # lc for coherence ***NOTE THESE OVERWRITE FUNCTION DEFS
_HAWC2_TURB_COOR = {'u': -1, 'v': -1, 'w': 1}  # hawc2 turb xyz to uvw


def check_chunk_idcs(freq, chunk_idcs):
    """Check chunk_idcs and update if it's not given"""
    return np.arange(np.size(freq)) if chunk_idcs is None else chunk_idcs


def check_sims_collocated(spat_df, con_tc):
    """Determine if the simulation points are all collocated with the 
    constraints. Returns boolean (True=all collocated)"""
    if con_tc is None:  # no constraints
        return False
    con_spat_df = con_tc.get_spat()
    # numerical arrays for which rows 
    con_arr = con_spat_df.T.values
    spat_arr = spat_df.T.values
    # magic with broadcasting (thank you numpy)
    sim_arr = spat_arr[:, np.newaxis, :]  # [ns, 1, 4] array of sim pts
    s_in_c = np.all(np.equal(sim_arr, con_arr), axis=2)  # [ns, nc] array of if si is in cj
    s_non_uniq = np.any(s_in_c, axis=1)  # [ns] array of if s is non unique
    return np.all(s_non_uniq)


def clean_turb(spat_df, all_spat_df, turb_df, decimals=10):
    """Remove the columns we don't return and rename the rest correctly. Will check only
    to decimals places when removing duplicates (data unchanged)."""
    # drop the columns that aren't in spat_df (use isclose for float comparisons)
    drop_cols = all_spat_df.apply(lambda col:
                                  not np.all(np.isclose(col.values, spat_df.values.T),
                                             axis=1).sum()).values
    turb_df.drop(turb_df.columns[drop_cols], axis=1, inplace=True)
    all_spat_df.drop(all_spat_df.columns[drop_cols], axis=1, inplace=True)
    # get unique locs in spat_df (need rounding logic here for dropping float duplicates)
    spat_xyz = spat_df.loc[['x', 'y', 'z']]
    spat_xyz = spat_xyz.T.loc[~spat_xyz.T.apply(np.round,
                                                args=[decimals]).duplicated()].T
    # for each column in all_spat_df, find the correct name using spat_df and rename it
    rename_map={}
    for colname in all_spat_df:
        col = all_spat_df[colname]
        k, x, y, z = col.values
        pid = np.all(np.isclose(np.array([x, y, z]), spat_xyz.values.T),
                     axis=1).argmax()  # use np.isclose for float comparisons
        new_name = f'{"uvw"[int(k)]}_p{pid}'
        if colname!=new_name:
            rename_map[colname]=new_name
    turb_df.rename(columns=rename_map, inplace=True)
    # order according to spat_df
    col_names = []
    for colname in spat_df:
        col = spat_df[colname]
        k, x, y, z = col.values
        pid = np.all(np.isclose(np.array([x, y, z]), spat_xyz.values.T),
                     axis=1).argmax()  # use np.isclose for float comparisons
        col_names.append(f'{"uvw"[int(k)]}_p{pid}')
    return turb_df[col_names]


def combine_spat_con(spat_df, con_tc, drop_duplicates=True, decimals=10):
    """Add constraining points in TimeConstraint to spat_df. NOTE constraints must come
    first or gen_turb will break. ALSO keep the data rows! Need them for corr. ALSO this
    function catches duplicates by first rounding to decimals places.
    """
    if con_tc is None:
        con_tc = TimeConstraint(index=_spat_rownames)
    con_spat_df = con_tc.get_spat().add_suffix('_con')
    if (set(spat_df.columns).intersection(set(con_spat_df.columns)) and (con_tc.size)):
        raise ValueError('Prohibited spat_df/con_tc column names! No column in spat_df' +
                         ' may have the same name as one on con_tc with "_con" ' +
                         'appended.')
    comb_df = pd.concat((con_spat_df, spat_df), axis=1)
    if drop_duplicates:
        # need to round before dropping to prevent machine-precision floats being "uniq"
        comb_df = comb_df.T.loc[~comb_df.T.apply(np.round,
                                                 args=[decimals]).duplicated()].T
    return comb_df


def gen_spat_grid(y, z, comps=[0, 1, 2]):
    """Generate spat_df (all turbulent components and grid defined by x and z)

    Notes
    -----
    0=u is downwind, 2=w is vertical and 1=v is lateral (right-handed
    coordinate system).
    """
    ys, zs = np.meshgrid(y, z)  # make a meshgrid
    ks = np.array(comps, dtype=int)  # sanitizing
    xs = np.zeros_like(ys)  # all points in a plane
    col_names = [f'{"uvw"[k]}_p{ip}' for ip in range(xs.size) for k in ks]
    spat_arr = np.c_[np.tile(comps, xs.size),
                     np.repeat(np.c_[xs.T.ravel(), ys.T.ravel(), zs.T.ravel()],
                               ks.size, axis=0)].T  # create array using numpy
    return pd.DataFrame(spat_arr, index=_spat_rownames, columns=col_names)


def get_chunk_idcs(freq, i_chunk, nf_chunk):
    """Return the range of indices corresponding to a specific chunk of frequencies."""
    n_f = len(freq)
    return np.arange(i_chunk*nf_chunk, min((i_chunk + 1)*nf_chunk, n_f))


def get_freq(**kwargs):
    """get frequency array"""
    t = np.arange(kwargs['nt']) * kwargs['T'] / kwargs['nt']
    n_f = kwargs['nt'] // 2 + 1
    freq = np.arange(n_f) / kwargs['T']
    return t, freq


def is_line(points):
    """see if the points given in points fall along a single line"""
    # first rotate by first pair to remove issue with vertical lines
    theta = np.arctan2((points[1, 1] - points[0, 1]), (points[1, 0] - points[0, 0]))
    R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    rot_points = points @ R
    # then do least-squares fit
    a = np.c_[rot_points[:, 0], np.ones(rot_points.shape[0])]
    b = rot_points[:, 1]
    x = np.linalg.lstsq(a, b, rcond=0)[0]
    if np.isclose(0., np.linalg.norm(a@x-b)):
        return True
    else:
        return False


def interpolator(points, values, xi):
    """Interpolate points in D dimensions.

    This function is based upon scipy's griddata function, but it has been expanded to
    handle single points and points that fall in a line. It is used in certain profile
    functions for the wind speed, turbulence standard deviation, and spectrum for
    interpolating the values calculated from ``con_tc`` to the new requested points.
    See examples of its usage in the Interpolator example.

    Parameters
    ----------
    points : ndarray of floats, shape (n, D)
        Data point coordinates. Can either be an array of
        shape (n, D), or a tuple of `D` arrays, each with shape `n`.
    values : ndarray of float or complex, shape (n,)
        Data values.
    xi : 2-D ndarray of float or tuple of 1-D array, shape (M, D)
        Points at which to interpolate data.

    Returns
    -------
    intp_val : ndarray
        Array of interpolated values, shape M.
    """
    ndim = 2  # function assumes we only have 2 interpolating dimensions (y and z)
    san_vars = [points, xi]  # variables we need to sanitize
    for i in range(len(san_vars)):
        if isinstance(san_vars[i], tuple):  # if it's a tuple, array it
            san_vars[i] = np.atleast_2d(san_vars[i])  # returns an ndim x np array
        elif isinstance(san_vars[i], (np.ndarray, list)):  # allow a list to be passed in
            san_vars[i] = np.atleast_2d(san_vars[i]).T  # returns an ndim x np array
        else:
            raise ValueError(f'Inputs points and xi can only be a tuple, list or numpy '
                             'array!')
        san_vars[i] = san_vars[i].T  # griddata takes np x ndim, so transpose
    points, xi = san_vars  # reassing points and xi
    # if one point, just return the values
    if (points.size == ndim):
        return sciint.griddata(points, values, xi, method='nearest')
    # if it's a line, use numpy's 1d interpolation
    if is_line(points):  # points lie in a line
        theta = np.arctan2((points[1, 1] - points[0, 1]), (points[1, 0] - points[0, 0]))
        R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
        rot_points, rot_xi = points @ R, xi @ R
        intp_val = np.interp(rot_xi[:, 0], rot_points[:, 0], values)
    else:  # more than 1 pt, not in a line (so 3+ points)
        intp_val = sciint.griddata(points, values, xi)  # 1st call griddata w/linear
        outside = np.isnan(intp_val)  # points outside grid will be nans
        out_intp = sciint.griddata(points, values, xi[outside],
                                   method='nearest')  # reeval outside grid with nearest
        intp_val[outside] = out_intp
    return intp_val


def make_hawc2_input(turb_dir, spat_df, **kwargs):
    """return strings for the hawc2 input files
    """
    # string of center position
    z_ref = kwargs['z_ref']
    str_cntr_pos0 = '  center_pos0             0.0 0.0 ' + \
        f'{-z_ref:.1f} ; hub height\n'

    # string for mann model block
    T, nt = kwargs['T'], kwargs['nt']
    y, z = set(spat_df.loc['y'].values), set(spat_df.loc['z'].values)
    n_x, du = nt, kwargs['u_ref'] * T / nt
    n_y, dv = len(y), (max(y) - min(y)) / (len(y) - 1)
    n_z, dw = len(z), (max(z) - min(z)) / (len(z) - 1)
    str_mann = '  begin mann ;\n' + \
               f'    filename_u {turb_dir}/u.bin ; \n' + \
               f'    filename_v {turb_dir}/v.bin ; \n' + \
               f'    filename_w {turb_dir}/w.bin ; \n' + \
               f'    box_dim_u {n_x:.0f} {du:.1f} ; \n' + \
               f'    box_dim_v {n_y:.0f} {dv:.1f} ; \n' + \
               f'    box_dim_w {n_z:.0f} {dw:.1f} ; \n' + \
               f'    dont_scale 1 ; \n' + \
               '  end mann '

    # string for output
    pts_df = spat_df.loc[['x', 'y', 'z'], spat_df.loc['k'] == 0]
    str_output = ''
    for col in pts_df.columns:
        x_p = -pts_df.loc['y', col]
        y_p = -pts_df.loc['x', col]
        z_p = -pts_df.loc['z', col]
        i_p = int(col.split('p')[-1])
        str_output += f'  wind free_wind 1 {x_p:.1f} {y_p:.1f} {z_p:.1f}' + \
            f' # wind_p{i_p} ; \n'

    return str_cntr_pos0, str_mann, str_output


def message(s, verbose=True):
    """Print a message to screen, if requested.
    """
    print(s) if verbose else None
    return


def rotate_time_series(ux, uy, uz):
    """Yaw and pitch time series so v- and w-directions have zero mean

        Args:
            ux (numpy array): array of x-sonic velocity
            uy (numpy array): array of y-sonic velocity
            uz (numpy array): array of z-sonic velocity

        Returns:
            x_rot (numpy array): [n_t x 3] array of rotated data (yaw+pitch)
            x_yaw (numpy array): [n_t x 3] array of rotated data (yaw)
    """

    # return all NaNs if any component is all nan values
    if all(np.isnan(ux)) * all(np.isnan(uy)) * all(np.isnan(uz)):
        u = np.zeros(ux.shape)
        v = np.zeros(uy.shape)
        w = np.zeros(uz.shape)
        u[:] = np.nan
        v[:] = np.nan
        w[:] = np.nan

    # if at least one data point in all three components
    else:

        # combine velocities into array
        x_raw = np.concatenate((ux.reshape(ux.size, 1),
                                uy.reshape(ux.size, 1),
                                uz.reshape(ux.size, 1)), axis=1)

        # interpolate out any NaN values
        for i_comp in range(x_raw.shape[1]):
            x = x_raw[:, i_comp]
            idcs_all = np.arange(x.size)
            idcs_notnan = np.logical_not(np.isnan(x))
            x_raw[:, i_comp] = np.interp(idcs_all, idcs_all[idcs_notnan], x[idcs_notnan])

        # rotate through yaw angle
        theta = np.arctan(np.nanmean(x_raw[:, 1]) / np.nanmean(x_raw[:, 0]))
        A_yaw = np.array([[np.cos(theta), -np.sin(theta), 0],
                          [np.sin(theta), np.cos(theta), 0],
                          [0, 0, 1]])
        x_yaw = x_raw @ A_yaw

        # rotate through pitch angle
        phi = np.arctan(np.nanmean(x_yaw[:, 2]) / np.nanmean(x_yaw[:, 0]))
        A_pitch = np.array([[np.cos(phi), 0, -np.sin(phi)],
                            [0, 1, 0],
                            [np.sin(phi), 0, np.cos(phi)]])
        x_rot = x_yaw @ A_pitch

        # if u is negative, we need to rotate 180 degrees around the y asix
        if x_rot[:, 0].sum() < 0:
            alpha = np.pi
            A_rot = np.array([[np.cos(alpha), 0, -np.sin(alpha)],
                              [0, 1, 0],
                              [np.sin(alpha), 0, np.cos(alpha)]])
            x_rot = x_rot @ A_rot

        # define rotated velocities
        u = x_rot[:, 0]
        v = x_rot[:, 1]
        w = x_rot[:, 2]

    return u, v, w



# -*- coding: utf-8 -*-
"""Setup file for PyConTurb

See README.md for how to use this file.
"""
from os import path
from setuptools import setup


def get_version():
    """Get version number from text file"""
    lines = open('./pyconturb/_version.py').readlines()
    version = [l for l in lines if '__version__' in l][0].split()[-1]
    return version.strip('"').strip("'")


def load_readme():
    """Load readme to put into pypi long description"""
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


setup(name='pyconturb',
      version=get_version(),
      description='An open-source constrained turbulence generator',
      long_description=load_readme(),
      long_description_content_type='text/markdown',
      url='https://gitlab.windenergy.dtu.dk/pyconturb/pyconturb',
      author='Jenni Rinker',
      author_email='rink@dtu.dk',
      license='MIT',
      packages=['pyconturb',  # top-level package
                ],
      install_requires=['h5py',  # load coherence array from file
                        'numpy',  # numberic arrays
                        'pandas',  # column-labelled arrays
                        'scipy',  # interpolating profile functions
                        ],
      zip_safe=False)



# -*- coding: utf-8 -*-
"""Base classes used in PyConTurb
"""
import numpy as np
import pandas as pd
from pandas import DataFrame


class TimeConstraint(DataFrame):
    """DataFrame-style object that specfies time constraints for simulation.

    The TimeConstraint object is essentially a pandas.DataFrame with a particular
    structure and some added useful methods. The first four rows correspond to ``k``,
    ``x``, ``y`` and ``z``. The remaining rows are time steps. Each column corresponds to
    a different location/turbulence component. The index must be of the form
    ``['k', 'x', 'y', 'z', <float array of time steps>]``. The columns need not have
    names.

    The object can be pickled/saved using same methods as pandas. To reload, load a
    properly-formatted pandas.DataFrame using the usual methods and pass it into the
    object: e.g.,``TimeConstraint(pd.read_csv(path))``.

    Parameters
    ----------
    data : ndarray (structured or homogeneous), Iterable, dict, or DataFrame.
        Dict can contain Series, arrays, constants, or list-like objects.
    index : Index or array-like
        Index to use for resulting frame. Will default to RangeIndex if
        no indexing information part of input data and no index provided
    columns : Index or array-like
        Column labels to use for resulting frame. Will default to
        RangeIndex (0, 1, 2, ..., n) if no column labels are provided
    dtype : dtype, default None
        Data type to force. Only a single dtype is allowed. If None, infer
    copy : boolean, default False
        Copy data from inputs. Only affects DataFrame / 2d ndarray input
    """

    def get_nt(self):
        """Return the numer of time steps in constraining data"""
        return self.get_time().shape[0]

    def get_spat(self):
        """Return a DataFrame with the spatial information"""
        return self[self.index.map(lambda x: type(x) is str)]

    def get_time(self):
        """Return a DataFrame with the spatial information"""
        return self[self.index.map(lambda x: type(x) is not str)]

    def get_T(self):
        """Return the total time of the constraining data"""
        t = self.get_time().index
        return t[-1] + t[1]

    def from_con_data(self, con_data=None, con_spat_df=None, con_turb_df=None):
        """Create TimeConstraint from old-style con_data dictionary"""
        # assing/check inputs
        if (con_data is None) and ((con_spat_df is None) or (con_turb_df is None)):
            raise ValueError('con_data was not given, so both con_spat_df and ' +
                             'con_turb_df must be given!')
        elif (con_data is not None):
            con_spat_df = con_data['con_spat_df']
            con_turb_df = con_data['con_turb_df']
        # check correct sizes
        if con_turb_df.shape[1] != con_spat_df.shape[0]:
            raise ValueError('No. rows in con_spat_df do not match columns in ' +
                             'con_turb_df!')
        try:
            con_spat_df = con_spat_df.T.drop('p_id').T
        except KeyError:
            pass
        con_df = pd.DataFrame(np.r_[con_spat_df.T, con_turb_df],
                              columns=con_turb_df.columns,
                              index=con_spat_df.columns.to_list()
                              + con_turb_df.index.to_list())
        self.__init__(con_df)
        return self



# -*- coding: utf-8 -*-
"""Define how the mean wind speed varies with y and z.

For example, the spatial variation of U could be a power law by height, which
is the wind profile specified in IEC 61400-1 Eds. 3 and 4. A user can also specify a
custom function to model the wind speed variation as a function of ``y`` and
``z``. See the notes in ``get_wsp_values`` below for more details.
"""
import warnings

import numpy as np

from pyconturb._utils import _DEF_KWARGS, interpolator


def get_wsp_values(spat_df, wsp_func, **kwargs):
    """Mean wind speed for points/components in ``spat_df``.

    The ``wsp_func`` must be a function of the form::

        wsp_values = wsp_func(spat_df, **kwargs)

    where ``spat_df`` is the spatial dataframe. You can use the
    profile functions built into PyConTurb (see below) or define your own
    custom functions. The output is assumed to be in m/s.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    wsp_func : function
        Function to map y and z to a wind speed in m/s.
    **kwargs
        Keyword arguments to pass into ``wsp_func``.

    Returns
    -------
    wsp_values : np.array
        [m/s] Mean wind speeds for the given spatial locations(s)/component(s).
        Dimension is ``(n_sp,)``.
    """
    wsp_values = wsp_func(spat_df, **kwargs)
    # wsp_values = wsp_func(spat_df.loc['y'], spat_df.loc['z'],
    #                       **kwargs) * (spat_df.loc['k'] == 0)
    return np.array(wsp_values)  # convert to array in case series given


def constant_profile(spat_df, u_ref=0, **kwargs):
    """Constant (or zero) mean wind speed on longitudinal turbulence components.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    u_ref : int/float, optional
        [m/s] Mean wind speed at all locations.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    wsp_values : np.array
        [m/s] Mean wind speeds at the specified location(s).
    """
    kwargs = {**{'u_ref': u_ref}, **kwargs}  # if not given, add to kwargs
    wsp_vals = np.ones_like(spat_df.loc['y']) * kwargs['u_ref']  # u, v, and w
    return wsp_vals * (spat_df.loc['k'] == 0)  # v, w set to zero


def data_profile(spat_df, con_tc=None, warn_datacon=True, **kwargs):
    """Mean wind speed interpolated from a TimeConstraint object.

    See the Examples and/or Reference Guide for details on the interpolator logic or for
    how to construct a TimeConstraint object.

    Note! If a component is requested for which there is no constraint, then this
    function will try to use the IEC spectra instead. Use the `warn_datacon` option to 
    disable the warning about this.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    con_tc : pyconturb.TimeConstraint
        [-] Constraint object. Must have correct format; see documentation on
        PyConTurb's TimeConstraint object for more details.
    warn_datacon : boolean
        [-] Warn if a requested component does not have a constraint, which results in
        an attempt at using the Kaimal spectrum. Default is True.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    wsp_values : np.array
        [m/s] Mean wind speed(s) at the specified location(s).
    """
    if con_tc is None:
        raise ValueError('No data provided!')
    u_mask = np.isclose(con_tc.loc['k'], 0)  # cols spat_df that correspond to u
    if not sum(u_mask):
        if warn_datacon:  # throw warning if requested
            warnings.warn('Longitudinal wind speed does not exist in constraints! '
                          + 'Cannot interpolate mean profile. Trying to use power law.',
                          Warning, stacklevel=2)
        return power_profile(spat_df, **kwargs)
    ypts = con_tc.filter(regex='u_').loc['y'].values.astype(float)
    zpts = con_tc.filter(regex='u_').loc['z'].values.astype(float)
    vals = con_tc.get_time().filter(regex='u_').mean().values.astype(float)
    y, z = spat_df.loc[['y', 'z']].values
    return interpolator((ypts, zpts), vals, (y, z))


def power_profile(spat_df, u_ref=_DEF_KWARGS['u_ref'], z_ref=_DEF_KWARGS['z_ref'],
                  alpha=_DEF_KWARGS['alpha'], **kwargs):
    """Power-law profile with height.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    u_ref : int/float, optional
        [m/s] Mean wind speed at reference height.
    z_ref : int/float, optional
        [m] Reference height.
    alpha : int/float, optional
        [-] Coefficient for the power law.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    wsp_values : np.array
        [m/s] Mean wind speed(s) at the specified location(s).
    """
    kwargs = {**{'u_ref': u_ref, 'z_ref': z_ref, 'alpha': alpha},
              **kwargs}  # if not given, add defaults to kwargs
    wsp_vals = kwargs['u_ref'] * (spat_df.loc['z'] / kwargs['z_ref']) ** kwargs['alpha']
    return wsp_vals * (spat_df.loc['k'] == 0)
